/*  This program is free software; you can redistribute it and/or modify
*  it under the terms of the GNU General Public License as published by
*  the Free Software Foundation; either version 2 of the License, or
*  (at your option) any later version.
*
*  This program is distributed in the hope that it will be useful,
*  but WITHOUT ANY WARRANTY; without even the implied warranty of
*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*  GNU General Public License for more details.
*  You should have received a copy of the GNU General Public License
*  along with this program; if not, write to the Free Software
*  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/
// (C) PicoPeta Simputers Pvt. Ltd. 

#include <stdio.h>
#include <stdlib.h>
#include <tk.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <string.h>
#include <X11/X.h>
#include "sched_sketch_core.h"

#define stip_wid 3
#define stip_ht 3

static sc_sk_node_t *new_sk_node(const char *id,
    unsigned short int x, unsigned short int y,
    unsigned short int width, unsigned int color,
    unsigned short int stipple);

static sc_stroke_node_t *new_stroke_node(signed char dx, signed char dy);
static void free_sk_node(sc_sk_node_t *skn);
static int do_pen_down( const char *id, unsigned short int x,
        unsigned short int y, unsigned short int width,
        unsigned int color, unsigned short int stipple);
static int do_pen_down_new( const char *id, unsigned short int x,
        unsigned short int y, unsigned short int width,
        unsigned int color, unsigned short int stipple, int draw_flg);
static int do_sk_add(signed char dx, signed char dy);
static int do_sk_add_new(short int x, short int y, int draw_flg);
static int do_sk_clean();
static void store_sk_node(sc_sk_node_t *skn, int fd);
static void sc_sk_draw_node(Tcl_Interp *interp, sc_sk_node_t *skn, 
        const char *pen_down_cb, const char *stroke_cb);
static void sc_sk_draw_node_new(Tcl_Interp *interp, sc_sk_node_t *skn);
static void dump_sk_node(sc_sk_node_t *skn);
static void set_line_attributes(Display *disp, Window win,
           unsigned short int width, unsigned int color,
           unsigned short int stipple);

// The global data structures
static sc_sk_node_t *sc_sk_head=NULL;
static sc_sk_node_t *sc_sk_tail=NULL;
// Clipboard linked list -- Subbu
static sc_sk_node_t *sc_sk_head_clipbrd=NULL;
static sc_sk_node_t *sc_sk_tail_clipbrd=NULL;

static char stip_bits[] = {0x02, 0x05, 0x02};
Pixmap sc_stip_map=None;
Display *sdisp;
Window swin;
GC sgc;
Tk_Window scanv, simgwin;
Colormap scmap;
XColor scol;
short int sx, sy;

void sc_sk_init ()
{
    sc_sk_head = NULL;
    sc_sk_tail = NULL;
    sc_sk_head_clipbrd = NULL;
    sc_sk_tail_clipbrd = NULL;
}

static sc_sk_node_t *new_sk_node(const char *id,
    unsigned short int x, unsigned short int y,
    unsigned short int width, unsigned int color,
    unsigned short int stipple)
{
    sc_sk_node_t *new_node;

    new_node = (sc_sk_node_t *)malloc(sizeof(sc_sk_node_t));
    if(!new_node) {
        fprintf(stderr,"%s(): Unable to allocate memory\n",__func__);
        return NULL;
    }
    new_node->x = x;
    new_node->y = y;
    new_node->width = width;
    new_node->color = color;
    new_node->stipple = stipple;
    new_node->next = NULL;
    new_node->prev = NULL;
    new_node->stroke_head = NULL;
    new_node->stroke_tail = NULL;
    if(strlen(id)<=ID_LEN)
        sprintf(new_node->id,"%10s",id);
    else
        strncpy(new_node->id,id,ID_LEN);
    new_node->id[ID_LEN]=0;
    return new_node;
}

static sc_stroke_node_t *new_stroke_node(signed char dx, signed char dy)
{
    sc_stroke_node_t *new_node;

    new_node = (sc_stroke_node_t *)malloc(sizeof(sc_stroke_node_t));
    if(!new_node) {
        fprintf(stderr,"%s(): Unable to allocate memory\n",__func__);
        return NULL;
    }
    new_node->no_of_pts = 0;
    memset(new_node->dx_dy, 0x00, 200);
    new_node->dx_dy[new_node->no_of_pts++] = dx;
    new_node->dx_dy[new_node->no_of_pts++] = dy;
    new_node->next = NULL;
    return new_node;
}

static void free_sk_node(sc_sk_node_t *skn) {
    // Free all nodes of the stroke buffer
    sc_stroke_node_t *stn;
    if(!skn)
        return;
    else
        stn=skn->stroke_head;

    while(stn) {
        skn->stroke_head = skn->stroke_head->next;
        free(stn);
        stn=skn->stroke_head;
    }
    free(skn);
}

static void sc_sk_draw_node(Tcl_Interp *interp, sc_sk_node_t *skn, 
        const char *pen_down_cb, const char *stroke_cb)
{
    char script[MAX_SCRIPT_SIZE];
    sc_stroke_node_t *stn;

    if(!skn)
        return;
    else
        stn = skn->stroke_head;

    sprintf(script,"%s %s %d %d %d %d %d",pen_down_cb,skn->id, skn->x, skn->y,
            skn->width, skn->color, skn->stipple);
    Tcl_Eval(interp,script);

    while(stn) {
        int i;
	for (i=0; i < stn->no_of_pts; i += 2) {
           sprintf(script,"%s %d %d",stroke_cb,stn->dx_dy[i],stn->dx_dy[i+1]);
           Tcl_Eval(interp,script);
        }
        stn=stn->next;
    }
}

static void sc_sk_draw_node_new(Tcl_Interp *interp, sc_sk_node_t *skn)
{
    sc_stroke_node_t *stn;

    if(!skn)
        return;
    else
        stn = skn->stroke_head;

    sx = skn->x; sy = skn->y;
    sgc = XCreateGC(sdisp, swin, 0, 0);
    set_line_attributes(sdisp, swin, skn->width, skn->color, skn->stipple);

    while (stn) {
        int i;
	for (i=0; i < stn->no_of_pts; i += 2) {
           XDrawLine(sdisp, swin, sgc, sx, sy, sx+(signed char)stn->dx_dy[i],
                    sy+(signed char)stn->dx_dy[i+1]);
           sx += stn->dx_dy[i]; sy += stn->dx_dy[i+1];
        }
        XFlush(sdisp);
        stn=stn->next;
    }
}

int sc_sk_draw(ClientData cd, Tcl_Interp *interp, int objc,
        Tcl_Obj *CONST objv[])
{
    char *id, *pen_down_cb, *stroke_cb;
    int do_all=0;
    char local_id[ID_LEN + 1];
    sc_sk_node_t *skn = sc_sk_head;

    if(objc!=4) {
        Tcl_WrongNumArgs(interp,1,objv,"<id> <pen_down_cb> <stroke_cb>");
        return TCL_ERROR;
    }

    if(!(id=Tcl_GetStringFromObj(objv[1],NULL)))
            return TCL_ERROR;
    if(!(pen_down_cb=Tcl_GetStringFromObj(objv[2],NULL)))
            return TCL_ERROR;
    if(!(stroke_cb=Tcl_GetStringFromObj(objv[3],NULL)))
            return TCL_ERROR;

    if(!strncmp(id,"all",3))
        do_all = 1;
    else {
        if(strlen(id)<=ID_LEN)
            sprintf(local_id,"%10s",id);
        else
            strncpy(local_id,id,ID_LEN);
        local_id[ID_LEN]=0;
    }

    while(skn) {
        if(do_all || !strcmp(local_id,skn->id))
            sc_sk_draw_node(interp, skn, pen_down_cb, stroke_cb);
        skn = skn->next;
    }
    return TCL_OK;

}

int sc_sk_draw_new(ClientData cd, Tcl_Interp *interp, int objc,
        Tcl_Obj *CONST objv[])
{
    char *id, *cnvs_win;
    int do_all=0;
    char local_id[ID_LEN + 1];
    sc_sk_node_t *skn = sc_sk_head;

    if(objc!=3) {
        Tcl_WrongNumArgs(interp,1,objv,"<id> <canv_name>");
        return TCL_ERROR;
    }

    if(!(id=Tcl_GetStringFromObj(objv[1],NULL)))
            return TCL_ERROR;
    if(!(cnvs_win=Tcl_GetStringFromObj(objv[2],NULL)))
            return TCL_ERROR;

    simgwin = Tk_MainWindow(interp);
    scanv = Tk_NameToWindow(interp, cnvs_win, simgwin);
    sdisp = Tk_Display(scanv);
    if (!sdisp) { return TCL_ERROR; }
    swin = Tk_WindowId(scanv);
    scmap = DefaultColormap(sdisp, 0);


    if ((sc_stip_map = XCreateBitmapFromData(sdisp, RootWindow(sdisp,
           DefaultScreen(sdisp)), stip_bits, stip_wid, stip_ht)) == 0) {
        perror("Can't create bitmap");
        return TCL_ERROR;
    }
    if(!strncmp(id,"all",3))
        do_all = 1;
    else {
        if(strlen(id)<=ID_LEN)
            sprintf(local_id,"%10s",id);
        else
            strncpy(local_id,id,ID_LEN);
        local_id[ID_LEN]=0;
    }

    while (skn) {
        if(do_all || !strcmp(local_id,skn->id))
            sc_sk_draw_node_new(interp, skn);
        skn = skn->next;
    }
    return TCL_OK;
}

int sc_sk_load(ClientData cd, Tcl_Interp *interp, int objc,
        Tcl_Obj *CONST objv[])
{
    char *file_name,*pen_down_cb,*stroke_cb;
    int fd;
    struct stat file_stat;
    off_t file_size;
    char *file_data=NULL, *tfile_data=NULL;
    char buffer[4096];
    int buff_size=4095;

    if(objc!=4) {
        Tcl_WrongNumArgs(interp,1,objv,"<filename> <pen_down_cb> <stroke_cb>");
        return TCL_ERROR;
    }
    if(!(file_name=Tcl_GetStringFromObj(objv[1],NULL)))
            return TCL_ERROR;

    if(!(pen_down_cb=Tcl_GetStringFromObj(objv[2],NULL)))
            return TCL_ERROR;

    if(!(stroke_cb=Tcl_GetStringFromObj(objv[3],NULL)))
            return TCL_ERROR;

    do_sk_clean();
    fd=open(file_name, O_RDONLY);
    if(fd < 0) {
	return TCL_ERROR;
    }
   
    if(fstat(fd, &file_stat)) {
	return TCL_ERROR;	
    }
   
    file_size = file_stat.st_size;

    if(file_size < 22) { return TCL_OK; }

    file_data = (char *)malloc(sizeof(char)*file_size);
    if(!file_data) {
	return TCL_ERROR;
    }

    tfile_data = file_data;

    if(read(fd, file_data, file_size) < file_size) {
	free(tfile_data);
	return TCL_ERROR;
    }


    buffer[0] = 0;

    while(file_size > 0) {
        unsigned short int x,y,width,stipple;
        unsigned int color;
        signed char dx,dy;
        char script[MAX_SCRIPT_SIZE];
        char id[ID_LEN + 1];

	memcpy(&id, file_data, ID_LEN);
	file_data += ID_LEN;
	file_size -= ID_LEN;

        id[ID_LEN]=0;

	memcpy(&x, file_data, 2);
	memcpy(&y, &(file_data[2]), 2);
	memcpy(&width, &(file_data[4]), 2);
	memcpy(&color, &(file_data[6]), 4);
	memcpy(&stipple, &(file_data[10]), 2);

	file_data += 12;
	file_size -= 12;

        do_pen_down(id,x,y,width,color,stipple);
        sprintf(script,"%s %s %d %d %d %d %d;",pen_down_cb,id,x,y,width,color,stipple);
	
	if(buff_size > strlen(script)) {
		strcat(buffer, script);
	} else {
        
		Tcl_Eval(interp,buffer);
		buffer[0] = 0;
		buff_size = 4095;
		strcat(buffer, script);
	}

	buff_size -= strlen(script);

	memcpy(&dx, file_data++, 1);
	file_size -= 1;

        while((dx != END_CHAR) && (file_size > 0)) {
	    memcpy(&dy, file_data++, 1);
            do_sk_add(dx, dy);
            sprintf(script,"%s %d %d;",stroke_cb,dx,dy);

            if(buff_size > strlen(script)) {
                 strcat(buffer, script);
            } else {
                 Tcl_Eval(interp,buffer);
                 buffer[0] = 0;
                 buff_size = 4095;
                 strcat(buffer, script);
            }
            buff_size -= strlen(script);

	    memcpy(&dx, file_data++, 1);
	    file_size -= 2;
        }
    }

    if(buff_size != 4095) {
	Tcl_Eval(interp, buffer);
    }

    close(fd);
    free(tfile_data);
    return TCL_OK;
}

int sc_sk_load_new(ClientData cd, Tcl_Interp *interp, int objc,
        Tcl_Obj *CONST objv[])
{
    char *file_name, *cnvs_win;
    char *file_data=NULL, *tfile_data=NULL;
    int fd;
    struct stat file_stat;
    off_t file_size;
    int draw_flg=1;

    if (objc!=3) {
        Tcl_WrongNumArgs(interp,1,objv,"<filename> <canv_name>");
        return TCL_ERROR;
    }
    if (!(file_name=Tcl_GetStringFromObj(objv[1],NULL)))
            return TCL_ERROR;

    if (!(cnvs_win=Tcl_GetStringFromObj(objv[2],NULL)))
            return TCL_ERROR;
    if (strcmp(cnvs_win,"dummy") == 0) {
       draw_flg = 0;
    } else {
       simgwin = Tk_MainWindow(interp);
       scanv = Tk_NameToWindow(interp, cnvs_win, simgwin);
       sdisp = Tk_Display(scanv);
       if (!sdisp) { return TCL_ERROR; }
       swin = Tk_WindowId(scanv);
       scmap = DefaultColormap(sdisp, 0);

       if ((sc_stip_map = XCreateBitmapFromData(sdisp, RootWindow(sdisp,
              DefaultScreen(sdisp)), stip_bits, stip_wid, stip_ht)) == 0) {
           perror("Can't create bitmap");
           return TCL_ERROR;
       }
    }
    do_sk_clean();
    fd=open(file_name, O_RDONLY);
    if (fd < 0) { return TCL_ERROR; }
   
    if (fstat(fd, &file_stat)) { return TCL_ERROR; }
    file_size = file_stat.st_size;
    if(file_size < 22) { return TCL_OK; }

    file_data = (char *)malloc(sizeof(char)*file_size);
    if(!file_data) {
	return TCL_ERROR;
    }

    tfile_data = file_data;

    if(read(fd, file_data, file_size) < file_size) {
	free(tfile_data);
	return TCL_ERROR;
    }

    while(file_size > 0) {
        unsigned short int x,y,width,stipple;
        unsigned int color;
        signed char dx,dy;
        char id[ID_LEN + 1];

	memcpy(&id, file_data, ID_LEN);
	file_data += ID_LEN;
	file_size -= ID_LEN;

        id[ID_LEN]=0;

	memcpy(&x, file_data, 2);
	memcpy(&y, &(file_data[2]), 2);
	memcpy(&width, &(file_data[4]), 2);
	memcpy(&color, &(file_data[6]), 4);
	memcpy(&stipple, &(file_data[10]), 2);

	file_data += 12;
	file_size -= 12;

        if (draw_flg) {
           sgc = XCreateGC(sdisp, swin, 0, 0);
           set_line_attributes(sdisp, swin, width, color, stipple);
        }

        do_pen_down_new(id, x, y, width, color, stipple, draw_flg);

	memcpy(&dx, file_data++, 1);
	file_size -= 1;

        while((dx != END_CHAR) && (file_size > 0)) {
	    memcpy(&dy, file_data++, 1);
            file_size--;
            do_sk_add_new(x+dx, y+dy, draw_flg);

            if (draw_flg) {
               XDrawLine(sdisp, swin, sgc, x, y, x+dx, y+dy);
            }
            x += dx; y += dy;

	    memcpy(&dx, file_data++, 1);
	    file_size--;
        }
        if (draw_flg) { XFlush(sdisp); }
    }
//    XFlush(sdisp);
    close(fd);
    free(tfile_data);
    return TCL_OK;
}

static void store_sk_node(sc_sk_node_t *skn, int fd)
{
    char end=END_CHAR;
    sc_stroke_node_t *stn = skn->stroke_head;
    char temp[2048];
    int count=22;

    memcpy(temp, &skn->id, 10);
    memcpy(&(temp[10]), &skn->x, 2);
    memcpy(&(temp[12]), &skn->y, 2);
    memcpy(&(temp[14]), &skn->width, 2);
    memcpy(&(temp[16]), &skn->color, 4);
    memcpy(&(temp[20]), &skn->stipple, 2);
//    write(fd, temp, 22);

    while(stn) {
        if ((count+stn->no_of_pts) > 2048) {
           write(fd,temp,count); count = 0;
        }
        memcpy(&(temp[count]), stn->dx_dy, (stn->no_of_pts+1));
        count += stn->no_of_pts;
	stn = stn->next;
    }

    memcpy(&(temp[count++]), &end, 1);
    write(fd, temp, count);

    //write(fd, &end, 1); // end of stroke marker
}

int sc_sk_store(ClientData cd, Tcl_Interp *interp, int objc,
        Tcl_Obj *CONST objv[])
{
    char *file_name;
    int fd;
    sc_sk_node_t *skn = sc_sk_head;

    if(objc!=2) {
        Tcl_WrongNumArgs(interp,1,objv,"<filename>");
        return TCL_ERROR;
    }
    if(!(file_name=Tcl_GetStringFromObj(objv[1],NULL)))
            return TCL_ERROR;

    fd=open(file_name,O_WRONLY|O_CREAT|O_TRUNC,0777);

    while(skn) {
        store_sk_node(skn, fd);
        skn = skn->next;
    }
    close(fd);
    return TCL_OK;
}

int sc_sk_delete(ClientData cd, Tcl_Interp *interp, int objc,
        Tcl_Obj *CONST objv[])
{
    char *id;
    char pid[ID_LEN+1];
    sc_sk_node_t *skn;

    if(objc!=2) {
        Tcl_WrongNumArgs(interp,1,objv,"<id>");
        return TCL_ERROR;
    }

    if(!(id=Tcl_GetStringFromObj(objv[1],NULL)))
            return TCL_ERROR;

    if(strlen(id)<=ID_LEN)
        sprintf(pid,"%10s",id);
    else
        strncpy(pid,id,ID_LEN);
    pid[ID_LEN]=0;

    skn = sc_sk_tail;

    while(skn) {
        if(!strncmp(skn->id,pid,ID_LEN)) {
            sc_sk_node_t *skt = skn;

            if(skn->prev) {
                skn->prev->next = skn->next;
            } else {
                sc_sk_head = skn->next;
                if(!sc_sk_head)
                    sc_sk_tail = NULL;
            }

            if(skn->next) {
                skn->next->prev = skn->prev;
            } else {
                sc_sk_tail = skn->prev;
                if(!sc_sk_tail)
                    sc_sk_head = NULL;
            }
            skn = skn->prev;
            free_sk_node(skt);
        } else {
            skn = skn->prev;
        }
    }

    return TCL_OK;
}

int sc_sk_copy(ClientData cd, Tcl_Interp * interp, int objc,
      Tcl_Obj *CONST objv[])
{
    char *sid=NULL, *did=NULL;
    char spid[ID_LEN+1], dpid[ID_LEN+1];
    sc_sk_node_t *skn=NULL;
    sc_stroke_node_t *tmpstn=NULL;

    if(objc!=3) {
        Tcl_WrongNumArgs(interp,1,objv,"<source-id> <dest-id>");
        return TCL_ERROR;
    }

    if(!(sid=Tcl_GetStringFromObj(objv[1],NULL)))
            return TCL_ERROR;
    if(!(did=Tcl_GetStringFromObj(objv[2],NULL)))
            return TCL_ERROR;
    if (!strcmp(sid,did)) { return TCL_ERROR; }

    if (strlen(sid) <= ID_LEN)
        sprintf(spid,"%10s",sid);
    else
        strncpy(spid,sid,ID_LEN);
    spid[ID_LEN]=0;
    if (strlen(did) <= ID_LEN)
        sprintf(dpid,"%10s",did);
    else
        strncpy(dpid,did,ID_LEN);
    dpid[ID_LEN]=0;

    skn = sc_sk_head;
    while (skn) {
        if(!strncmp(skn->id,spid,ID_LEN)) {
            sc_sk_node_t *tmpskn = new_sk_node(dpid, skn->x, skn->y,
                                         skn->width,skn->color,skn->stipple);
            sc_sk_tail->next = tmpskn;
            tmpskn->prev = sc_sk_tail;
            sc_sk_tail = tmpskn;
            tmpstn = skn->stroke_head;
            while (tmpstn) {
               sc_stroke_node_t *stn = 
                         new_stroke_node(tmpstn->dx_dy[0], tmpstn->dx_dy[1]);
               if(sc_sk_tail->stroke_head == NULL) {
                  sc_sk_tail->stroke_head = stn;
                  sc_sk_tail->stroke_tail = sc_sk_tail->stroke_head;
               } else {
                  sc_sk_tail->stroke_tail->next = stn;
                  sc_sk_tail->stroke_tail = stn;
               }
               stn->no_of_pts = tmpstn->no_of_pts;
               memcpy(stn->dx_dy, tmpstn->dx_dy, tmpstn->no_of_pts);
               tmpstn = tmpstn->next;
            }
        }
        skn = skn->next;
    }
    return TCL_OK;
}

int sc_sk_merge(ClientData cd, Tcl_Interp *interp, int objc,
        Tcl_Obj *CONST objv[])
{
    char *id1, *id2;
    char pid1[ID_LEN+1];
    char pid2[ID_LEN+1];
    sc_sk_node_t *skn;

    if(objc!=3) {
        Tcl_WrongNumArgs(interp,1,objv,"<id1> <id2>");
        return TCL_ERROR;
    }

    if(!(id1=Tcl_GetStringFromObj(objv[1],NULL)))
            return TCL_ERROR;

    if(!(id2=Tcl_GetStringFromObj(objv[2],NULL)))
            return TCL_ERROR;

    if(strlen(id1)<=ID_LEN)
        sprintf(pid1,"%10s",id1);
    else
        strncpy(pid1,id1,ID_LEN);
    pid1[ID_LEN]=0;

    if(strlen(id2)<=ID_LEN)
        sprintf(pid2,"%10s",id2);
    else
        strncpy(pid2,id2,ID_LEN);
    pid2[ID_LEN]=0;

    skn = sc_sk_head;

    while(skn) {
        if(!strncmp(skn->id,pid2,ID_LEN)) {
            strcpy(skn->id,pid1);
        }
        skn = skn->next;
    }

    return TCL_OK;
}

int sc_sk_undo(ClientData cd, Tcl_Interp *interp, int objc,
        Tcl_Obj *CONST objv[])
{
    char *id;
    char pid[ID_LEN+1];
    sc_sk_node_t *skn;
    char found=0;

    if(objc!=2) {
        Tcl_WrongNumArgs(interp,1,objv,"<id>");
        return TCL_ERROR;
    }

    if(!(id=Tcl_GetStringFromObj(objv[1],NULL)))
            return TCL_ERROR;

    if(strlen(id)<=ID_LEN)
        sprintf(pid,"%10s",id);
    else
        strncpy(pid,id,ID_LEN);
    pid[ID_LEN]=0;

    skn = sc_sk_tail;

    while(skn && !found) {
        if(!strncmp(skn->id,pid,ID_LEN)) {
            found=1;
            break;
        }
        skn = skn->prev;
    }

    if(found) {
        if(skn->prev) {
            skn->prev->next = skn->next;
        } else {
            sc_sk_head = skn->next;
            if(!sc_sk_head)
                sc_sk_tail = NULL;
        }

        if(skn->next) {
            skn->next->prev = skn->prev;
        } else {
            sc_sk_tail = skn->prev;
            if(!sc_sk_tail)
                sc_sk_head = NULL;
        }

        free_sk_node(skn);
    }

    return TCL_OK;
}

int sc_sk_undo_new(ClientData cd, Tcl_Interp *interp, int objc,
        Tcl_Obj *CONST objv[])
{
    char *id;
    char pid[ID_LEN+1];
    sc_sk_node_t *skn;
    char found=0;

    if(objc!=2) {
        Tcl_WrongNumArgs(interp,1,objv,"<id>");
        return TCL_ERROR;
    }

    if(!(id=Tcl_GetStringFromObj(objv[1],NULL)))
            return TCL_ERROR;

    if(strlen(id)<=ID_LEN)
        sprintf(pid,"%10s",id);
    else
        strncpy(pid,id,ID_LEN);
    pid[ID_LEN]=0;

    skn = sc_sk_tail;

    while(skn && !found) {
        if(!strncmp(skn->id,pid,ID_LEN)) {
            found=1;
            break;
        }
        skn = skn->prev;
    }

    if(found) {
        if(skn->prev) {
            skn->prev->next = skn->next;
        } else {
            sc_sk_head = skn->next;
            if(!sc_sk_head)
                sc_sk_tail = NULL;
        }

        if(skn->next) {
            skn->next->prev = skn->prev;
        } else {
            sc_sk_tail = skn->prev;
            if(!sc_sk_tail)
                sc_sk_head = NULL;
        }

        free_sk_node(skn);
    }

    XClearWindow(sdisp, swin);
    return TCL_OK;
}

void set_line_attributes(Display *disp, Window win,
           unsigned short int width, unsigned int color,
           unsigned short int stipple)
{
   char str[10];

   sprintf(str, "#%06x", color);
   XParseColor(disp, scmap, str, &scol);
   XAllocColor(disp, scmap, &scol);
   XSetForeground(disp, sgc, scol.pixel);
   XSetLineAttributes(disp, sgc, width, LineSolid, CapRound, JoinRound);
   if (stipple) {
      if (sc_stip_map != None)
         XFreePixmap(disp, sc_stip_map);
      if ((sc_stip_map = XCreateBitmapFromData(disp, win, stip_bits,
                     stip_wid, stip_ht)) == 0)
         return;
      XSetStipple(disp, sgc, sc_stip_map);
      XSetFillStyle(disp, sgc, FillStippled);
   } else {
      XSetFillStyle(disp, sgc, FillSolid);
   }
}

// backend for sk_pen_down
static int do_pen_down(const char *id, unsigned short int x,
        unsigned short int y, unsigned short int width,
        unsigned int color, unsigned short int stipple)
{
    if (sc_sk_head == NULL) {
        sc_sk_head = new_sk_node(id, x, y, width, color, stipple);
        sc_sk_tail = sc_sk_head;
    } else {
        sc_sk_node_t *skn = new_sk_node(id, x, y, width, color, stipple);
        sc_sk_tail->next = skn;
        skn->prev = sc_sk_tail;
        sc_sk_tail = skn;
    }

    return TCL_OK;
}

// backend for sk_pen_down_new
static int do_pen_down_new(const char *id, unsigned short int x,
        unsigned short int y, unsigned short int width,
        unsigned int color, unsigned short int stipple, int draw_flg)
{
    if (sc_sk_head == NULL) {
        sc_sk_head = new_sk_node(id, x, y, width, color, stipple);
        sc_sk_tail = sc_sk_head;
    } else {
        sc_sk_node_t *skn = new_sk_node(id, x, y, width, color, stipple);
        sc_sk_tail->next = skn;
        skn->prev = sc_sk_tail;
        sc_sk_tail = skn;
    }
    sx = x; sy = y;
    if (draw_flg) {
       sgc = XCreateGC(sdisp, swin, 0, 0);
       set_line_attributes(sdisp, swin, width, color, stipple);
    }
    return TCL_OK;
}

int sc_sk_pen_down(ClientData cd, Tcl_Interp *interp, int objc,
        Tcl_Obj *CONST objv[])
{
    int x,y,width,color,stipple;
    char *id;

    if(objc!=7) {
        Tcl_WrongNumArgs(interp,1,objv,"<id> <x> <y> <width> <color> <stipple>");
        return TCL_ERROR;
    }

    if(!(id=Tcl_GetStringFromObj(objv[1],NULL)))
        return TCL_ERROR;

    if(Tcl_GetIntFromObj(interp,objv[2],&x)!=TCL_OK)
        return TCL_ERROR;

    if(Tcl_GetIntFromObj(interp,objv[3],&y)!=TCL_OK)
        return TCL_ERROR;

    if(Tcl_GetIntFromObj(interp,objv[4],&width)!=TCL_OK)
        return TCL_ERROR;

    if(Tcl_GetIntFromObj(interp,objv[5],&color)!=TCL_OK)
        return TCL_ERROR;

    if(Tcl_GetIntFromObj(interp,objv[6],&stipple)!=TCL_OK)
        return TCL_ERROR;

    return do_pen_down( (const char *)id,
                        (unsigned short int)x,
                        (unsigned short int)y,
                        (unsigned short int)width,
                        (unsigned int)color,
                        (unsigned short int)stipple);
}

int sc_sk_pen_down_new(ClientData cd, Tcl_Interp *interp, int objc,
        Tcl_Obj *CONST objv[])
{
    int x,y,width,color,stipple;
    char *id, *cnvs_win;

    if(objc!=8) {
        Tcl_WrongNumArgs(interp,1,objv,"<id> <x> <y> <width> <color> <stipple> <canv_name>");
        return TCL_ERROR;
    }

    if(!(id=Tcl_GetStringFromObj(objv[1],NULL)))
        return TCL_ERROR;

    if(Tcl_GetIntFromObj(interp,objv[2],&x)!=TCL_OK)
        return TCL_ERROR;

    if(Tcl_GetIntFromObj(interp,objv[3],&y)!=TCL_OK)
        return TCL_ERROR;

    if(Tcl_GetIntFromObj(interp,objv[4],&width)!=TCL_OK)
        return TCL_ERROR;

    if(Tcl_GetIntFromObj(interp,objv[5],&color)!=TCL_OK)
        return TCL_ERROR;

    if(Tcl_GetIntFromObj(interp,objv[6],&stipple)!=TCL_OK)
        return TCL_ERROR;

    if(!(cnvs_win=Tcl_GetStringFromObj(objv[7],NULL)))
        return TCL_ERROR;

    simgwin = Tk_MainWindow(interp);
    scanv = Tk_NameToWindow(interp, cnvs_win, simgwin);
    sdisp = Tk_Display(scanv);
    if (!sdisp) { return TCL_ERROR; }
    swin = Tk_WindowId(scanv);
    scmap = DefaultColormap(sdisp, 0);

    return do_pen_down_new( (const char *)id,
                        (unsigned short int)x,
                        (unsigned short int)y,
                        (unsigned short int)width,
                        (unsigned int)color,
                        (unsigned short int)stipple, 1);
}

static int do_sk_add(signed char dx, signed char dy)
{
    //fprintf(stderr,__func__"(), dx=%d,dy=%d\n",(int)dx,(int)dy);
    if(!sc_sk_tail || !sc_sk_head)
        return TCL_ERROR;
    if(sc_sk_tail->stroke_head == NULL) {
        sc_sk_tail->stroke_head = new_stroke_node(dx, dy);
        sc_sk_tail->stroke_tail = sc_sk_tail->stroke_head;
    } else {
        sc_stroke_node_t *stn = sc_sk_tail->stroke_tail;
	if (stn->no_of_pts < 200) {
           stn->dx_dy[stn->no_of_pts++] = dx;
           stn->dx_dy[stn->no_of_pts++] = dy;
        } else {
           sc_stroke_node_t *stn = new_stroke_node(dx, dy);
           sc_sk_tail->stroke_tail->next = stn;
           sc_sk_tail->stroke_tail = stn;
        }
    }
    return TCL_OK;
}

static int do_sk_add_new(short int x, short int y, int draw_flg)
{
    if(!sc_sk_tail || !sc_sk_head)
        return TCL_ERROR;
    if(sc_sk_tail->stroke_head == NULL) {
        sc_sk_tail->stroke_head = new_stroke_node((signed char) (x-sx),
                                                  (signed char) (y-sy));
        sc_sk_tail->stroke_tail = sc_sk_tail->stroke_head;
    } else {
        sc_stroke_node_t *stn = sc_sk_tail->stroke_tail;
	if (stn->no_of_pts < 200) {
           stn->dx_dy[stn->no_of_pts++] = (signed char) (x-sx);
           stn->dx_dy[stn->no_of_pts++] = (signed char) (y-sy);
        } else {
           sc_stroke_node_t *stn = new_stroke_node((signed char)(x-sx),
                                                   (signed char)(y-sy));
           sc_sk_tail->stroke_tail->next = stn;
           sc_sk_tail->stroke_tail = stn;
        }
    }
    if (draw_flg) {
       XDrawLine(sdisp, swin, sgc, sx, sy, x, y);
    }
    sx = x; sy = y;
    return TCL_OK;
}

int sc_sk_add(ClientData cd, Tcl_Interp *interp, int objc,
        Tcl_Obj *CONST objv[])
{
    int dx,dy;

    if(objc!=3) {
        Tcl_WrongNumArgs(interp,1,objv,"<dx> <dy>");
        return TCL_ERROR;
    }

    if(Tcl_GetIntFromObj(interp,objv[1],&dx)!=TCL_OK)
        return TCL_ERROR;

    if(Tcl_GetIntFromObj(interp,objv[2],&dy)!=TCL_OK)
        return TCL_ERROR;

    //fprintf(stderr,__func__"() , dx=%d,dy=%d\n",dx,dy);
    return do_sk_add((signed char)dx,(signed char)dy);
}

int sc_sk_add_new(ClientData cd, Tcl_Interp *interp, int objc,
        Tcl_Obj *CONST objv[])
{
    int x,y;

    if(objc!=3) {
        Tcl_WrongNumArgs(interp,1,objv,"<x> <y>");
        return TCL_ERROR;
    }

    if(Tcl_GetIntFromObj(interp,objv[1],&x)!=TCL_OK)
        return TCL_ERROR;

    if(Tcl_GetIntFromObj(interp,objv[2],&y)!=TCL_OK)
        return TCL_ERROR;

    //fprintf(stderr,__func__"() , dx=%d,dy=%d\n",dx,dy);
    return do_sk_add_new(x,y,1);
}

// Complete cleanup
static int do_sk_clean()
{
    sc_sk_node_t *skn = sc_sk_head;

    //fprintf(stderr,__func__"()\n");
    while(skn) {
        sc_sk_head = sc_sk_head->next;
        free_sk_node(skn);
        skn = sc_sk_head;
    }
    sc_sk_head = sc_sk_tail = NULL;
    return TCL_OK;
}

int sc_sk_clean(ClientData cd, Tcl_Interp *interp, int objc,
        Tcl_Obj *CONST objv[])
{
    return do_sk_clean();
}

int sc_sk_clean_new(ClientData cd, Tcl_Interp *interp, int objc,
        Tcl_Obj *CONST objv[])
{
    char *cnvs_win=NULL;

    if (objc != 2) {
       Tcl_WrongNumArgs(interp, 1, objv, "<canv_name>");
       return TCL_ERROR;
    }
    if (!(cnvs_win=Tcl_GetStringFromObj(objv[1],NULL)))
       return TCL_ERROR;

    simgwin = Tk_MainWindow(interp);
    scanv = Tk_NameToWindow(interp, cnvs_win, simgwin);
    sdisp = Tk_Display(scanv);
    if (!sdisp) { return TCL_ERROR; }
    swin = Tk_WindowId(scanv);
    XClearWindow(sdisp, swin);

    return do_sk_clean();
}

static void dump_sk_node(sc_sk_node_t *skn)
{
    sc_stroke_node_t *stn;

    if(!skn)
        return;
    else
        stn=skn->stroke_head;

    fprintf(stderr,"id=%s,x=%d,y=%d,width=%d,color=%d,stipple=%d\n",skn->id,skn->x,skn->y,skn->width,skn->color,skn->stipple);

    while(stn) {
        int i;
        for (i=0; i < stn->no_of_pts; i += 2) {
           fprintf(stderr,"\tdx=%d,dy=%d\n",stn->dx_dy[i],stn->dx_dy[i+1]);
        }
        stn=stn->next;
    }
}

// Debug dump
int sc_sk_dump(ClientData cd, Tcl_Interp *interp, int objc,
        Tcl_Obj *CONST objv[])
{
    sc_sk_node_t *skn = sc_sk_head;

    //fprintf(stderr,__func__"()\n");
    while(skn) {
        dump_sk_node(skn);
        skn = skn->next;
    }
    return TCL_OK;
}

int sc_sk_moveto_clipbrd(ClientData cd, Tcl_Interp *interp, int objc,
			Tcl_Obj *CONST objv[])
{
    char *id;
    char pid[ID_LEN+1];
    sc_sk_node_t *skn, *tmpskn;
    char found=0;

    if(objc!=2) {
        Tcl_WrongNumArgs(interp,1,objv,"<id>");
        return TCL_ERROR;
    }

    if(!(id=Tcl_GetStringFromObj(objv[1],NULL)))
            return TCL_ERROR;

    if(strlen(id)<=ID_LEN)
        sprintf(pid,"%10s",id);
    else
        strncpy(pid,id,ID_LEN);
    pid[ID_LEN]=0;

    skn = sc_sk_head;

    while(skn) {
        tmpskn = skn->next;
        if(!strncmp(skn->id,pid,ID_LEN)) {
            if(skn->prev) {
                skn->prev->next = skn->next;
            } else {
                sc_sk_head = skn->next;
                if(!sc_sk_head)
                    sc_sk_tail = NULL;
            }
            if(skn->next) {
                skn->next->prev = skn->prev;
            } else {
                sc_sk_tail = skn->prev;
                if(!sc_sk_tail)
                    sc_sk_head = NULL;
            }
            skn->next = NULL; skn->prev = NULL;
            if (sc_sk_head_clipbrd) {
               sc_sk_tail_clipbrd->next = skn;
               skn->prev = sc_sk_tail_clipbrd;
               sc_sk_tail_clipbrd = skn;
            } else {
               sc_sk_head_clipbrd = skn; sc_sk_tail_clipbrd = skn;
            }
        }
        skn = tmpskn;
    }

    return TCL_OK;
}

int sc_sk_movefrom_clipbrd(ClientData cd, Tcl_Interp * interp, int objc,
      Tcl_Obj *CONST objv[])
{
    char *id, *newid;
    char pid[ID_LEN+1], newpid[ID_LEN+1];
    sc_sk_node_t *skn, *tmp_skn;

    if(objc!=3) {
        Tcl_WrongNumArgs(interp,1,objv,"<id> <new-id>");
        return TCL_ERROR;
    }

    if(!(id=Tcl_GetStringFromObj(objv[1],NULL)))
            return TCL_ERROR;

    if(!(newid=Tcl_GetStringFromObj(objv[2],NULL)))
            return TCL_ERROR;

    if(strlen(id)<=ID_LEN)
        sprintf(pid,"%10s",id);
    else
        strncpy(pid,id,ID_LEN);
    pid[ID_LEN]=0;
    if(strlen(newid)<=ID_LEN)
        sprintf(newpid,"%10s",newid);
    else
        strncpy(newpid,newid,ID_LEN);
    newpid[ID_LEN]=0;

    skn = sc_sk_head_clipbrd;

    while(skn) {
        sc_sk_node_t *tmpskn = skn->next;
        if(!strncmp(skn->id,pid,ID_LEN)) {
            strcpy(skn->id,newpid);
            if(skn->prev) {
                skn->prev->next = skn->next;
            } else {
                sc_sk_head_clipbrd = skn->next;
                if(!sc_sk_head_clipbrd)
                    sc_sk_tail_clipbrd = NULL;
            }
            if(skn->next) {
                skn->next->prev = skn->prev;
            } else {
                sc_sk_tail_clipbrd = skn->prev;
                if(!sc_sk_tail_clipbrd)
                    sc_sk_head_clipbrd = NULL;
            }
            if (sc_sk_head) {
               sc_sk_tail->next = skn;
               skn->prev = sc_sk_tail;
               sc_sk_tail = skn;
            } else {
               sc_sk_head = skn; sc_sk_tail = skn;
            }
        }
        skn = tmpskn;
    }

    return TCL_OK;
}

int sc_sk_copyto_clipbrd(ClientData cd, Tcl_Interp * interp, int objc,
      Tcl_Obj *CONST objv[])
{
    char *id=NULL;
    char pid[ID_LEN+1];
    sc_sk_node_t *skn=NULL;
    sc_stroke_node_t *tmpstn=NULL;

    if(objc!=2) {
        Tcl_WrongNumArgs(interp,1,objv,"<id>");
        return TCL_ERROR;
    }

    if(!(id=Tcl_GetStringFromObj(objv[1],NULL)))
            return TCL_ERROR;

    if(strlen(id)<=ID_LEN)
        sprintf(pid,"%10s",id);
    else
        strncpy(pid,id,ID_LEN);
    pid[ID_LEN]=0;

    skn = sc_sk_head;

    while (skn) {
        if(!strncmp(skn->id,pid,ID_LEN)) {
            if (sc_sk_head_clipbrd == NULL) {
               sc_sk_head_clipbrd = 
                     new_sk_node(skn->id,skn->x,skn->y,skn->width,
                                      skn->color,skn->stipple);
               sc_sk_tail_clipbrd = sc_sk_head_clipbrd;
            } else {
               sc_sk_node_t *tmpskn = new_sk_node(skn->id,skn->x,skn->y,
                                         skn->width,skn->color,skn->stipple);
               sc_sk_tail_clipbrd->next = tmpskn;
               tmpskn->prev = sc_sk_tail_clipbrd;
               sc_sk_tail_clipbrd = tmpskn;
            }
            if(!sc_sk_tail_clipbrd || !sc_sk_head_clipbrd)
               return TCL_ERROR;
            tmpstn = sc_sk_tail_clipbrd->stroke_head;
            while (tmpstn) {
               sc_stroke_node_t *stn = 
                         new_stroke_node(tmpstn->dx_dy[0], tmpstn->dx_dy[1]);
               if(sc_sk_tail_clipbrd->stroke_head == NULL) {
                  sc_sk_tail_clipbrd->stroke_head = stn;
                  sc_sk_tail_clipbrd->stroke_tail = 
                         sc_sk_tail_clipbrd->stroke_head;
               } else {
                  sc_sk_tail_clipbrd->stroke_tail->next = stn;
                  sc_sk_tail_clipbrd->stroke_tail = stn;
               }
               stn->no_of_pts = tmpstn->no_of_pts;
               memcpy(stn->dx_dy, tmpstn->dx_dy, tmpstn->no_of_pts);
               tmpstn = tmpstn->next;
            }
        }
        skn = skn->next;
    }
    return TCL_OK;
}

int sc_sk_merge_clipbrd(ClientData cd, Tcl_Interp *interp, int objc,
        Tcl_Obj *CONST objv[])
{
    char *id1, *id2;
    char pid1[ID_LEN+1];
    char pid2[ID_LEN+1];
    sc_sk_node_t *skn;

    if(objc!=3) {
        Tcl_WrongNumArgs(interp,1,objv,"<id1> <id2>");
        return TCL_ERROR;
    }

    if(!(id1=Tcl_GetStringFromObj(objv[1],NULL)))
            return TCL_ERROR;

    if(!(id2=Tcl_GetStringFromObj(objv[2],NULL)))
            return TCL_ERROR;

    if(strlen(id1)<=ID_LEN)
        sprintf(pid1,"%10s",id1);
    else
        strncpy(pid1,id1,ID_LEN);
    pid1[ID_LEN]=0;

    if(strlen(id2)<=ID_LEN)
        sprintf(pid2,"%10s",id2);
    else
        strncpy(pid2,id2,ID_LEN);
    pid2[ID_LEN]=0;

    skn = sc_sk_head_clipbrd;

    while(skn) {
        if(!strncmp(skn->id,pid2,ID_LEN)) {
            strcpy(skn->id,pid1);
        }
        skn = skn->next;
    }
    return TCL_OK;
}

static int do_sk_clean_clipbrd()
{
    sc_sk_node_t *skn = sc_sk_head_clipbrd;

    //fprintf(stderr,__func__"()\n");
    while(skn) {
        sc_sk_head_clipbrd = sc_sk_head_clipbrd->next;
        free_sk_node(skn);
        skn = sc_sk_head_clipbrd;
    }
    return TCL_OK;
}

int sc_sk_clean_clipbrd(ClientData cd, Tcl_Interp *interp, int objc,
        Tcl_Obj *CONST objv[])
{
    return do_sk_clean_clipbrd();
}

int sc_sk_draw_rect(ClientData cd, Tcl_Interp *interp, int objc,
        Tcl_Obj *CONST objv[])
{
    char *cnvs_win;
    int x1, y1, x2, y2, width, color;

    if(objc!=8) {
        Tcl_WrongNumArgs(interp,1,objv,"<x1> <y1> <x2> <y2> <width> <color> <canv_name>");
        return TCL_ERROR;
    }

    if(Tcl_GetIntFromObj(interp,objv[1],&x1)!=TCL_OK)
        return TCL_ERROR;
    if(Tcl_GetIntFromObj(interp,objv[2],&y1)!=TCL_OK)
        return TCL_ERROR;
    if(Tcl_GetIntFromObj(interp,objv[3],&x2)!=TCL_OK)
        return TCL_ERROR;
    if(Tcl_GetIntFromObj(interp,objv[4],&y2)!=TCL_OK)
        return TCL_ERROR;
    if(Tcl_GetIntFromObj(interp,objv[5],&width)!=TCL_OK)
        return TCL_ERROR;
    if(Tcl_GetIntFromObj(interp,objv[6],&color)!=TCL_OK)
        return TCL_ERROR;
    if(!(cnvs_win=Tcl_GetStringFromObj(objv[7],NULL)))
            return TCL_ERROR;

    simgwin = Tk_MainWindow(interp);
    scanv = Tk_NameToWindow(interp, cnvs_win, simgwin);
    sdisp = Tk_Display(scanv);
    if (!sdisp) { return TCL_ERROR; }
    swin = Tk_WindowId(scanv);
    scmap = DefaultColormap(sdisp, 0);

    sgc = XCreateGC(sdisp, swin, 0, 0);
    set_line_attributes(sdisp, swin, width, color, 0);

    XDrawLine(sdisp, swin, sgc, x1, y1, x2, y1);
    XDrawLine(sdisp, swin, sgc, x2, y1, x2, y2);
    XDrawLine(sdisp, swin, sgc, x2, y2, x1, y2);
    XDrawLine(sdisp, swin, sgc, x1, y2, x1, y1);
    return TCL_OK;
}

int sc_sk_draw_lines(ClientData cd, Tcl_Interp *interp, int objc,
        Tcl_Obj *CONST objv[])
{
    char *cnvs_win;
    int x1, y1, x2, y2, spacing, color;
    int y;

    if(objc!=8) {
        Tcl_WrongNumArgs(interp,1,objv,"<x1> <y1> <x2> <y2> <spacing> <color> <canv_name>");
        return TCL_ERROR;
    }

    if(Tcl_GetIntFromObj(interp,objv[1],&x1)!=TCL_OK)
        return TCL_ERROR;
    if(Tcl_GetIntFromObj(interp,objv[2],&y1)!=TCL_OK)
        return TCL_ERROR;
    if(Tcl_GetIntFromObj(interp,objv[3],&x2)!=TCL_OK)
        return TCL_ERROR;
    if(Tcl_GetIntFromObj(interp,objv[4],&y2)!=TCL_OK)
        return TCL_ERROR;
    if(Tcl_GetIntFromObj(interp,objv[5],&spacing)!=TCL_OK)
        return TCL_ERROR;
    if(Tcl_GetIntFromObj(interp,objv[6],&color)!=TCL_OK)
        return TCL_ERROR;
    if(!(cnvs_win=Tcl_GetStringFromObj(objv[7],NULL)))
            return TCL_ERROR;

    simgwin = Tk_MainWindow(interp);
    scanv = Tk_NameToWindow(interp, cnvs_win, simgwin);
    sdisp = Tk_Display(scanv);
    if (!sdisp) { return TCL_ERROR; }
    swin = Tk_WindowId(scanv);
    scmap = DefaultColormap(sdisp, 0);

    sgc = XCreateGC(sdisp, swin, 0, 0);
    set_line_attributes(sdisp, swin, 1, color, 0);

    for (y=y1+spacing; y < y2; y += spacing) {
        XDrawLine(sdisp, swin, sgc, x1, y, x2, y);
    }
    return TCL_OK;
}

int sc_sk_clear(ClientData cd, Tcl_Interp *interp, int objc,
        Tcl_Obj *CONST objv[])
{
    char *cnvs_win=NULL;

    if (objc != 2) {
       Tcl_WrongNumArgs(interp, 1, objv, "<canv_name>");
       return TCL_ERROR;
    }
    if (!(cnvs_win=Tcl_GetStringFromObj(objv[1],NULL)))
       return TCL_ERROR;

    simgwin = Tk_MainWindow(interp);
    scanv = Tk_NameToWindow(interp, cnvs_win, simgwin);
    sdisp = Tk_Display(scanv);
    if (!sdisp) { return TCL_ERROR; }
    swin = Tk_WindowId(scanv);
    XClearWindow(sdisp, swin);

    return TCL_OK;
}
